*  Program...........: RunFox.PRG
*  Author............: Edward Leafe
*  Project...........: RunFox; runtime command window utility
*  Created...........: 07/10/93
*
*  Copyright.........: (c) 1993-94 by Edward Leafe
*
*) Description.......: A utility program that allows the use of most of the
*                      functionality of the native FoxPro Command Window
*                      without the full FoxPro program being resident on the
*                      system. It requires the usual runtime libraries that
*                      allow for the distribution of compiled FoxPro .EXE's.
*
*  Calling Samples...: (from DOS): RUNFOX
*                      (from a FoxPro app): ON KEY LABEL F12 DO RUNFOX WITH .T.
*
*  Parameter List....: tlSaveView - Flag for saving/restoring original view
*
*  Major change list.: 08/14/93 - Added error handling routine.
*                      09/29/93 - Save screen image; Output to screen instead of window;
*                        browse window is hidden; removed WAIT WIND on output;
*                        added DIR to output commands; added F2 toggle for hiding all windows
*                      01/04/94 - Ported to FoxPro 2.5 DOS and Windows
*
* Cost...............: This program is shareware. You may use it for 30 days
*                      without charge. If, after 30 days you decide that you
*                      don't like or can't use the program, simply delete it
*                      from your computer. If you decide to include RunFox in
*                      any distributed application, or wish to continue using
*                      it after 30 days, the cost is $13.00 US. Registered
*                      users may include RunFox royalty-free in their
*                      distributed applications. Software registration fees
*                      should be sent to:
*                            Edward Leafe
*                            212 Wilson St.
*                            Hackensack, NJ 07601-3046
*
* Support............: The best way to reach me is through CompuServe mail.
*                      My address on CIS is: 72530,1175. Any ideas, suggestions
*                      for improvements or (shudder) bug reports are welcome.
*
* Distribution.......: You may freely distribute this program in its entirety
*                      without modification; any enhancements or changes may
*                      be included, but they must be marked as such, and the
*                      original, unmodified files must be included. You may
*                      not charge for the distribution of this program, nor
*                      include it with any collection of software for which a
*                      fee is charged, without the written authorization of
*                      the author. User groups which charge a nominal fee for
*                      the cost of diskettes are specifically excluded from
*                      this restriction.
*

**********************************
****** S E T U P   C O D E  ******
**********************************

PARAMETER tlSaveView, tlSaveScrn
PRIVATE ALL LIKE j*
PRIVATE ALL LIKE s*
PUSH KEY CLEAR
IF SET("TALK") = "ON"
   SET TALK OFF
   scTalk = "ON"
ELSE
   scTalk = "OFF"
ENDIF
tlSaveView = !EMPTY(tlSaveView)
tlSaveScrn = !EMPTY(tlSaveScrn)
IF tlSaveView
   scViewFile = "RUNFOX.VUE"
   CREATE VIEW (scViewFile)
   scAlias = ALIAS()
   snSelect = SELECT()
ENDIF
scWoutput = WOUTPUT()
IF tlSaveScrn
   IF !EMPTY(scWoutput)
      SAVE SCREEN TO ssWoutput
   ENDIF
   ACTI SCREEN
   SAVE SCREEN TO ssScreen
ENDIF
IF !EMPTY(scWoutput)
   ACTI WIND (scWoutput)
ENDIF

*** NOTE: Due to the way Fox handles variable scoping, you need to define
* any variables to be used in the RunFox command window here. I have included
* all the single-letter variable names, so that RunFox commands such as
*   COUNT TO X FOR !DELETED
*   ?X
* will work. If you are more comfortable with other variable names, add or
* replace the ones below with the names you wish to use. Alternatively,
* declare them PUBLIC first; RunFox will track them and release them upon exit.
*
*** NOTE: again, due to scoping limitations, commands like 'SCATTER MEMVAR'
* will not work as expected. I know of no direct workaround, unless you
* declare every memvar PUBLIC first, which would limit the usefulness
* of the program. Avoid memvar use; instead, use one of the pre-defined
* variables below as follows:
*    DIMENSION X[1,1]
*    SCATTER TO X
*    INSERT INTO YourDBF FROM X
*
* Initialize variables for temporary usage
PRIVATE A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z
STORE "" TO A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z
* Initialize array to hold 'local' public variables
PRIVATE laPublic, lnPublic
DIMENSION laPublic[1]
lnPublic = 0                                     && Counts how many 'publics' have been declared

scExact = SET("EXACT")
SET EXACT OFF
scSetANSI = SET("ANSI")
SET ANSI OFF
scMenu = MENU()
IF !EMPTY(scMenu)
   PUSH MENU &scMenu
ENDIF
SET SYSMENU TO DEFAULT
SET SYSMENU AUTOMATIC
scClock = SET("CLOCK")
DO CASE
   CASE "FOXPRO 2.0" $ UPPER(VERSION(1)) OR "FOXPRO/LAN 2.0" $ UPPER(VERSION(1)) OR _DOS
      SET CLOCK ON
   CASE _WINDOWS
      SET CLOCK STATUS
ENDCASE
scCursor = SET("CURSOR")
SET CURSOR ON                                    && Make sure cursor is visible
scError = ON("ERROR")                            && Save the ON ERROR routine
ON ERROR DO RFError WITH ERROR(), MESSAGE()
scSetEsc = SET("ESCAPE")
SET ESCAPE OFF
ON KEY LABEL ESC DO CmdClear
* Provide a means of hiding all windows, to mimic ALT-CTRL-SHIFT
ON KEY LABEL F2 DO CmdToggl
* Provide a shortcut for quitting RunFox
ON KEY LABEL ALT+Q DO GetOut                     && Hot Key to Quit

PRIVATE lcCR, lcLF, lcCRLF, lcSemi, lcMark
DO CASE
   CASE "FOXPRO 2.0" $ UPPER(VERSION(1)) OR "FOXPRO/LAN 2.0" $ UPPER(VERSION(1)) OR _DOS
      lcMark = CHR(255)
   CASE _WINDOWS
      lcMark = CHR(160)                          && Prints blank in MS Sans Serif
ENDCASE
lcCR   = CHR(13)
lcLF   = CHR(10)
lcCRLF = CHR(13) + CHR(10)
lcSemi = CHR(59)                                 && Semi-colon
PRIVATE lcCmdHist, lcLastCmd
lcCmdHist = ""                                   && Internal storage of command window history
lcLastCmd = ""                                   && Last command entered

* Define the window for browsing
DO CASE
   CASE "FOXPRO 2.0" $ UPPER(VERSION(1)) OR "FOXPRO/LAN 2.0" $ UPPER(VERSION(1)) OR _DOS
      DEFINE WINDOW wbRFBrowse ;
      FROM 1, 0 ;
      TO 24,79 ;
      FLOAT ;
      CLOSE ;
      SHADOW ;
      MINIMIZE ;
      SYSTEM ;
      COLOR SCHEME 8
   CASE _WINDOWS
      DEFINE WINDOW wbRFBrowse ;
      AT 0.000, 0.000 ;
      SIZE 25.00,100.000 ;
      FONT "MS Sans Serif", 8 ;
      STYLE "B" ;
      FLOAT ;
      CLOSE ;
      SHADOW ;
      MINIMIZE ;
      SYSTEM
ENDCASE


******************************************************************************
DO RunFox.SPR                                    && Call the import screen
******************************************************************************

**********************************
*** C L E A N - U P   C O D E  ***
**********************************

RELEASE WINDOW wbRFBrowse
* Release all the local 'public' variables, if any
IF lnPublic > 0
   FOR jnCnt = 1 TO lnPublic
      jcVar = laPublic[jnCnt]
      RELEASE (jcVar)
   ENDFOR
ENDIF

* Restore screen and windows
IF !EMPTY(scWoutput)
   ACTI WIND (scWoutput)
   IF tlSaveView
      RESTORE SCREEN FROM ssWoutput
   ENDIF
ENDIF
IF tlSaveScrn
   ACTI SCREEN
   RESTORE SCREEN FROM ssScreen
ENDIF
* Restore the original view, if required
IF tlSaveView
   SET VIEW TO (scViewFile)
   ERASE (scViewFile)
   IF !EMPTY(scAlias)
      SELECT (scAlias)
   ELSE
      SELECT (snSelect)
   ENDIF
ENDIF
IF scCursor = "OFF"
   SET CURSOR OFF
ENDIF
IF scSetEsc = "ON"
   SET ESCAPE ON
ENDIF
IF scExact = "ON"
   SET EXACT ON
ENDIF
IF scSetANSI = "ON"
   SET ANSI ON
ENDIF
SET CLOCK &scClock
IF !EMPTY(scMenu)
   POP MENU &scMenu
ENDIF
ON ERROR &scError
IF scTalk = "ON"
   SET TALK ON
ENDIF
POP KEY
RETURN


PROCEDURE RunIt
   PRIVATE ALL LIKE j*
   PRIVATE ALL LIKE s*
   jnLen     = LEN(lcProg)
   jnLine    = 1                                 && Which line are we on?
   jnRetPos  = AT(lcCRLF,lcProg,jnLine)          && Find the first CR-LF
   jnLastPos = 1                                 && Starting position for SUBSTR command
   jlFound   = .F.
   * Get the position of the mark character
   jnMarkPos = AT(lcMark,lcProg)
   * Find out how many CRLF's are prior to that position
   jnPrevRets = StrCnt(SUBSTR(lcProg,1,jnMarkPos),lcCRLF)
   * Find out how many CRLF's are in the whole command window
   jnTotRets = StrCnt(lcProg,lcCRLF)
   IF jnTotRets > jnPrevRets
      IF jnPrevRets = 0
         jnPos1 = 1
      ELSE
         jnPos1 = AT(lcCRLF,lcProg,jnPrevRets) + 2
      ENDIF
      jnPos2 = AT(lcCRLF,lcProg,jnPrevRets+1)
      jcLine = SUBSTR(lcProg,jnPos1,(jnPos2-jnPos1))
   ELSE
      IF jnPrevRets = 0                          && First line
         jcLine = lcProg
      ELSE
         jnPos1 = AT(lcCRLF,lcProg,jnPrevRets) + 2
         jcLine = SUBSTR(lcProg,jnPos1)
      ENDIF
   ENDIF

   jlOKtoRun = .T.                               && Flag for complete command
   jcLine = ALLTRIM(jcLine)
   jcLine = STRTRAN(jcLine,lcMark)               && Remove the marker

   * Check if last character is semi-colon
   IF RIGHT(jcLine,1) = lcSemi
      * Need to check for subsequent lines (if any)
      jnCurrRet = jnPrevRets + 1
      DO WHILE jnTotRets > jnCurrRet
         jnPos1 = AT(lcCRLF,lcProg,jnCurrRet) + 2
         jnPos2 = AT(lcCRLF,lcProg,jnCurrRet+1)
         jcNextLine = SUBSTR(lcProg,jnPos1,(jnPos2-jnPos1))
         jcNextLine = ALLTRIM(jcNextLine)
         jcLine = jcLine + jcNextLine
         IF RIGHT(jcNextLine,1) # lcSemi         && No more continuation lines
            EXIT
         ENDIF
         jnCurrRet = jnCurrRet + 1
      ENDDO
      IF RIGHT(jcLine,1) = lcSemi
         * Last line is continued; don't run it yet
         jlOKtoRun = .F.
      ENDIF
   ENDIF
   * If the end of the command is finished and there are previous lines,
   * we need to see if any previous lines are continuation lines
   IF jlOKtoRun AND (jnPrevRets > 0)
      jnCurrRet = jnPrevRets -1
      DO WHILE jnCurrRet >= 0
         IF jnCurrRet = 0
            jnPos1 = 1
         ELSE
            jnPos1 = AT(lcCRLF,lcProg,jnCurrRet) + 2
         ENDIF
         jnPos2 = AT(lcCRLF,lcProg,jnCurrRet+1)
         jcPrevLine = SUBSTR(lcProg,jnPos1,(jnPos2-jnPos1))
         jcPrevLine = ALLTRIM(jcPrevLine)
         IF RIGHT(jcPrevLine,1) = lcSemi         && Prev. line was a continuation
            jcLine = jcPrevLine + jcLine
         ELSE
            EXIT
         ENDIF
         jnCurrRet = jnCurrRet - 1
      ENDDO
   ENDIF

   * Add it to the end of lcCmdHist if it is different than the
   * last command. NOTE: jcLine and lcLastCmd are compared to each other
   * twice to account for the way XBase handles string comparison when
   * SET EXACT is OFF
   ** Need to add CR-LF to all but the last semi-colon to make the string
   ** appear correctly in the command window
   jnSemiCnt = StrCnt(jcLine,lcSemi)
   IF jnSemiCnt > 0
      IF RIGHT(jcLine,1) = lcSemi                && Don't add CR-LF to that one
         jcLine = SUBSTR(jcLine,1,LEN(jcLine)-1)  && Strip the last semi-colon
         jcLine = STRTRAN(jcLine,lcSemi,lcSemi+lcCRLF)
         jcLine = jcLine + lcSemi
      ELSE
         jcLine = STRTRAN(jcLine,lcSemi,lcSemi+lcCRLF)
      ENDIF
   ENDIF
   IF jlOKtoRun AND ;
      ((lcLastCmd # jcLine) OR (jcLine # lcLastCmd) OR EMPTY(lcCmdHist))
      lcCmdHist = lcCmdHist + jcLine + lcCRLF
      * Filter out blank lines
      DO WHILE StrCnt(lcCmdHist,(lcCRLF+lcCRLF)) > 0
         lcCmdHist = STRTRAN(lcCmdHist,(lcCRLF+lcCRLF),lcCRLF)
      ENDDO
      lcProg = lcCmdHist
   ELSE
      lcProg = STRTRAN(lcProg,lcMark) + lcCRLF
      * Filter out blank lines
      DO WHILE StrCnt(lcProg,(lcCRLF+lcCRLF)) > 0
         lcProg = STRTRAN(lcProg,(lcCRLF+lcCRLF),lcCRLF)
      ENDDO
   ENDIF
   IF jlOKtoRun
      lcLastCmd = jcLine
   ENDIF

   * Now clean up jcLine for execution
   jcLine = STRTRAN(jcLine,lcSemi)               && Strip the semi-colons
   jcLine = STRTRAN(jcLine,lcCR)                 && Strip the carriage returns
   jcLine = STRTRAN(jcLine,lcLF)                 && Strip the line feeds
   jcLine = ALLTRIM(jcLine)
   * Check for empty command, i.e., pressing ENTER on a blank line
   IF EMPTY(jcLine)
      lcProg = lcCmdHist
      jlOKtoRun = .F.
   ENDIF

   * See if the user wishes to exit
   jlExit = .F.
   IF INLIST(UPPER(LEFT(jcLine,4)),"EXIT","CANC","QUIT")
      DO GetOut
   ENDIF

   * See if the user wishes to clear the Command Window
   IF INLIST(UPPER(jcLine),"CLEAR COMM","CLEA COMM","CLEAR CMD","CLEA CMD")
      STORE "" TO lcProg, lcCmdhist              && Clear the previous commands
      jlOKtoRun = .F.
      SHOW GET lcProg
   ENDIF

   * See if the user has declared a local 'public' variable
   jnSpcPos = AT(" ",jcLine)
   IF jnSpcPos > 0
      jcCmd = ALLTRIM(UPPER(SUBSTR(jcLine,1,jnSpcPos-1)))
      IF INLIST(jcCmd,"PUBL","PUBLIC")           && User is declaring a variable
         jcVarStr = SUBSTR(jcLine,jnSpcPos+1)    && Get the rest of the line
         jnCommaCnt = StrCnt(jcVarStr,",")       && See how many commas are in the line
         DO WHILE jnCommaCnt > 0
            jnCommaPos = AT(",",jcVarStr)
            jcVar = SUBSTR(jcVarStr,1,jnCommaPos -1)  && Get the part before the comma
            jcVarStr = SUBSTR(jcVarStr,jnCommaPos+1)  && Get the part after the comma
            IF ASCAN(laPublic,jcVar) = 0         && Variable not already in array
               lnPublic = lnPublic + 1
               DIMENSION laPublic[lnPublic]      && Re-dimension the array of publics
               laPublic[lnPublic] = jcVar        && Add the variable name to the array
            ENDIF
         ENDDO
         IF ASCAN(laPublic,jcVarStr) = 0            && Variable not already in array
            lnPublic = lnPublic + 1
            DIMENSION laPublic[lnPublic]         && Re-dimension the array of publics
            laPublic[lnPublic] = jcVarStr        && Add the variable name to the array
         ENDIF
      ENDIF
   ENDIF

   * Screen out illegal commands
   * NOTE: this can be fleshed out a bit more
   IF INLIST(UPPER(LEFT(jcLine,7)),"DO CASE","DO WHIL") OR ;
      INLIST(UPPER(LEFT(jcLine,4)),"FOR ","SCAN")
      DO RFError WITH 0, "That command is not allowed in ;
      interactive mode"
      jlOKtoRun = .F.
   ENDIF
   IF UPPER(LEFT(jcLine,9)) = "CLEAR ALL" OR ;
      UPPER(LEFT(jcLine,8)) = "CLEA ALL"
      DO RFError WITH 0, "That command will crash the program!" + ;
      lcCRLF + "Don't do it!!"
      jlOKtoRun = .F.
   ENDIF

   IF jlOKtoRun
      * Make any browse commands behave normally, i.e., don't browse in the
      * small command window and save their images on the screen
      IF INLIST(UPPER(jcLine),"BROW","BROWSE","EDIT")
         jcClause = " SAVE"
         IF NOT (" LAST" $ UPPER(jcLine))        && Add the 'Normal' option
            jcClause = jcClause + " NORMAL"
         ENDIF
         IF WEXIST("wbRFBrowse")
            jcClause = jcClause + " WINDOW wbRFBrowse"
         ENDIF
         jcLine = jcLine + jcClause
      ENDIF

      * Check for commands that require screen output
      IF LEFT(jcLine,1) = "?" OR INLIST(UPPER(LEFT(jcLine,4)),"DISP","LIST","CLEA","DIR")
         ACTIVATE SCREEN
      ENDIF
      * Run the line through SpltMcro in case it is too long
      PUSH KEY CLEAR                             && Remove the ON KEY LABEL ENTER effect
      DO SpltMcro WITH jcLine
      POP KEY
      IF UPPER(WOUTPUT()) # "WRRUNFOX"
         ACTIVATE WINDOW wrRunFox
      ENDIF
   ENDIF

   IF !jlExit
      ** Calculate how many lines to the end of the text
      jnDown = (StrCnt(lcProg,lcCRLF) - jnLine) + 1  && Number of lines from bottom of lcProg
      IF UPPER(WOUTPUT()) = "WRRUNFOX"
         KEYBOARD REPLICATE("{DNARROW}",jnDown*2) PLAIN  && Add some extras for good measure
      ENDIF
   ENDIF
RETURN


PROCEDURE SpltMcro
   * Routine to take a long macro line
   * & split it into pieces so that it can be run
   * It also runs the macro.
   * (c) Flash Creative Management
   PARAMETERS tcMacro
   PRIVATE ALL LIKE j*
   PRIVATE ALL LIKE s*
   PRIVATE laMacro
   IF LEN(tcMacro) > 2048
      ?? CHR(7)
      WAIT WINDOW "Command length too long" TIMEOUT 2
   RETURN
ENDIF
* Calculate # of splits that need to be done
jnNumSplit = INT(LEN(tcMacro)/254)
IF LEN(tcMacro) % 254 > 0
   jnNumSplit = jnNumSplit + 1
ENDIF
DIMENSION laMacro[jnNumSplit]
lcRun = ''
FOR i = 1 TO jnNumSplit
   laMacro[i] = SUBSTR(tcMacro,1+((i-1)*254),254)
   lcRun = lcRun+"&"+"laMacro["+STR(i,2)+"]"
ENDFOR
* Execute the command
&lcRun
RETURN


PROCEDURE CmdToggl
   * This procedure alternately hides all windows or shows all windows.
   * It is analogous to pressing ALT-CTRL-SHIFT in regular FoxPro
   IF WVISIBLE("wrRunFox")
      HIDE WIND ALL
      HIDE MENU ALL
   ELSE
      SHOW MENU ALL
      SHOW WIND ALL SAME
      HIDE WIND wbRFBrowse                       && Keep this window hidden
   ENDIF
RETURN


PROCEDURE DoWhen
   *** This procedure runs the WHEN clause of a field.
   PARAMETERS tcWhichOne,tlFakeKey
   IF PARAMETERS() < 2
      tlFakeKey = .F.
   ENDIF
   IF PARAMETERS() < 1
      tcWhichOne = VARREAD()
   ENDIF
   tcWhichOne = UPPER(tcWhichOne)
   PRIVATE ALL LIKE j*
   jlRetVal = .T.
   DO CASE
      CASE tcWhichOne = UPPER("Read")
         HIDE WIND wbRFBrowse
   ENDCASE
RETURN jlRetVal


PROCEDURE CmdClear
   PRIVATE ALL LIKE j*
   PRIVATE ALL LIKE s*
   IF UPPER(WONTOP()) = "WRRUNFOX"
      * Need to restore command window
      lcProg = lcCmdHist
      SHOW GET lcProg
      KEYBOARD REPLICATE("{DNARROW}",20) PLAIN
   ELSE
      KEYBOARD "{ESC}" PLAIN
   ENDIF
RETURN


PROCEDURE GetOut
   CLEAR READ
   jlOKtoRun = .F.
   jlExit = .T.
RETURN

